;*****************************************************************************/
;
;	PROJECT:	PSOC MORSE DECODER
; 	FILE NAME: 	MORSEDECODER.ASM
; 	AUTHOR:		Melchor Varela
;
; 	DESCRIPTION
;
;  	Morse decoder.
;
; 	HISTORY
;
;	NAME   	DATE		REMARKS
;
;	MVM	   	1APR2003	Creation
; 	MVM		11OCT2003	Tone detection and debouncing implemented outside 
;						this module. Added protection for WPM table read.
;
;*****************************************************************************/
;*****************************************************************************/
;
;	$Workfile:   morsedecoder.asm  $
;	$Author:   mvm  $
;	$Date:   Oct 11 2003 21:24:04  $
;	$Revision:   1.3  $
;
;*****************************************************************************/

include 'm8c.inc'
include 'lcd.inc'

	export 	_MorseDecoder

area 	bss(RAM)

;-----------------------------------------------------------------------------
;  Defines
;-----------------------------------------------------------------------------
DIT_DEFAULT:		equ	14			; Default values for 20WPM
DAH_DEFAULT:		equ	36	
THRESHOLD_DEFAULT:	equ	(DAH_DEFAULT + DIT_DEFAULT)/2

NB_LEARN_PERIODS:	equ	5			; Idle time to restart learning: 5*255*5ms = 6seg
NUM_SAMPLES_TONE:	equ	50			; First level debounce 

MINIMUM_THRES:		equ	10
;-----------------------------------------------------------------------------
;  Private data: 
;-----------------------------------------------------------------------------
ditptr:		BLK		1		; Pointer to DIT buffer
ditvals:	BLK		4		; DIT buffer
ditave:		BLK		1		; Average of DIT periods
ditsum:		BLK		2		; Sum of last four DIT periods

dahptr:		BLK		1		; Pointer to DASH buffer
dahvals:	BLK		4		; DASH buffer
dahave:		BLK		1		; Average of DASH periods
dahsum:		BLK		2		; Sum of last four DASH periods

threshold:	BLK		1		; Symbol detection threshold	
period:		BLK		1		; Gets last period measurement
character:	BLK		1		; Last decoded character
code:		BLK		1		; Variables for decoding: code mask
mask:		BLK		1		; ""
linebuffer:	BLK		16+1	; Display line buffer for decoding data
tmp:		BLK		1		; Temporary variable to help 16bit arithmetic
column:		BLK		1		; Current display cursor position for decoding data
	
cntlearn:	BLK		1		; Auxiliary counter to measure idle time and start learning

area 	text(ROM,REL)

;-----------------------------------------------------------------------------
;  FUNCTION NAME: MorseDecoder
;
;  DESCRIPTION:
;     Morse decoder main program. It never returns.
;
;  ARGUMENTS:
;     none.
;
;  RETURNS:
;     none.
;
;  SIDE EFFECTS:
;     none.
;
;-----------------------------------------------------------------------------
_MorseDecoder:

	; Clears line buffer buffer
	mov		[column], 0
	mov		X, 16
   	mov		[X+linebuffer], 0
	mov		X, 15
eraseLoop:	
   	mov		[X+linebuffer], 0x20
   	dec		X
   	jnc		eraseLoop	

	lcall	ClearLine1
	
	call	DisplaySilence	; To erase row 1, column 0
	
restartLearning:
	call	DisplayStatusMsg
	
	; Init decoder variables
	mov		[cntlearn], 0
	
	mov		[code], 0		
	mov		[mask], 1
	
	mov 	[threshold], THRESHOLD_DEFAULT
	mov		[ditptr], 0
	mov		[dahptr], 0
		
	mov		[ditave], 	 DIT_DEFAULT
	mov		[ditvals+0], DIT_DEFAULT
	mov		[ditvals+1], DIT_DEFAULT
	mov		[ditvals+2], DIT_DEFAULT
	mov		[ditvals+3], DIT_DEFAULT
	mov		[ditsum],    4*DIT_DEFAULT
	mov		[ditsum+1],  0
	
	mov		[dahave], 	 DAH_DEFAULT
	mov		[dahvals+0], DAH_DEFAULT
	mov		[dahvals+1], DAH_DEFAULT
	mov		[dahvals+2], DAH_DEFAULT
	mov		[dahvals+3], DAH_DEFAULT
	mov		[dahsum],    4*DAH_DEFAULT
	mov		[dahsum+1],  0
	
	mov		[_Counter5ms], 0

	; Main decoding loop
mainLoop:		 
	call	WaitTone
	cmp		[period], 255					; Received tone ?
	jnz		analyseSymbol					; Jumps if tone received
							
	cmp		[mask], 1						; Character reception in course ?
	jnz		analyseSymbol
	inc		[cntlearn]
	cmp		[cntlearn], NB_LEARN_PERIODS	; Idle time completed ?
	jnc		restartLearning		
	jmp		mainLoop

analyseSymbol:	
	mov		[cntlearn], 0					; Resets learning control counter
	
	cmp		[mask], 1		
	jz		analyseTone						; Jumps if not processing in course
	
	mov		A, [period]						; Period contains the last pause duration
	cmp		A, [threshold]	
	jc		analyseTone						; Jumps if period < threshold
											; Symbol processing done
	mov		A, [mask]
	add		[code], A
	call	TranslateCharacter
	call	DisplayCharacter
	mov		[code], 0						; Inits decoding variables
	mov		[mask], 1
	mov		A, [threshold]
	asl		A
	cmp		A, [period]						; period > (threshold*2)
	jnc		testGotoMainLoop			
	mov		[character], 0x20				; Detected word space
	call	DisplayCharacter

testGotoMainLoop:
	cmp		[period], 255					; Received tone ?
	jz		mainLoop
	
analyseTone:	
	call	WaitSilence						; Measure tone length

	mov		A, [period]
	cmp		A, [threshold]	
	jc		isADit							; dit == period < threshold
	call	CalcAveDah						; DASH symbol received
	jmp		isAxxx_End
isADit:
	call	CalcAveDit						; DIT symbol received
	mov		A, [mask]
	add		[code], A
isAxxx_End:			
	asl		[mask]
	jmp		mainLoop		
	
	
;-----------------------------------------------------------------------------
;  FUNCTION NAME: CalcAveDit
;
;  DESCRIPTION:
;     Computes the DIT moving average and update threshold
;
;  ARGUMENTS:
;     [period]		Contains the last DIT period measurement
;
;  RETURNS:
;     [threshold]	Updated with the mean value of DIT and DASH periods	
;
;  SIDE EFFECTS:
;     none.
;
;-----------------------------------------------------------------------------
CalcAveDit:
	mov		X, [ditptr]					; Calculate moving average and update buffer	
	mov		A, [X+ditvals]
	sub		[ditsum], A
	sbb		[ditsum+1], 0		

	mov		A, [period]
	mov		[X+ditvals], A
	add		[ditsum], A
	adc		[ditsum+1], 0
	
	inc		[ditptr]
	and		[ditptr], 0x3

	mov		A, [ditsum]					; Computes average
	mov		[tmp], [ditsum+1]
	asr		[tmp]
	rrc		A
	asr		[tmp]
	rrc		A
	mov		[ditave], A

	mov		[tmp], 0					; Update threshold
	mov		A, [dahave]
	add		A, [ditave]
	adc		[tmp], 0	
	asr		[tmp]
	rrc		A
	mov		[threshold], A

	ret	
;-----------------------------------------------------------------------------
;  FUNCTION NAME: CalcAveDah
;
;  DESCRIPTION:
;     Computes the DASH moving average and update threshold
;
;  ARGUMENTS:
;     [period]		Contains the last DASH period measurement
;
;  RETURNS:
;     [threshold]	Updated with the mean value of DIT and DASH periods	
;
;  SIDE EFFECTS:
;     none.
;
;-----------------------------------------------------------------------------	
CalcAveDah:
	mov		X, [dahptr]				; Calculate moving average and update buffer	
	mov		A, [X+dahvals]
	sub		[dahsum], A
	sbb		[dahsum+1], 0		

	mov		A, [period]
	mov		[X+dahvals], A
	add		[dahsum], A
	adc		[dahsum+1], 0
	
	inc		[dahptr]
	and		[dahptr], 0x3

	mov		A, [dahsum]				; Computes average
	mov		[tmp], [dahsum+1]
	asr		[tmp]
	rrc		A
	asr		[tmp]
	rrc		A
	mov		[dahave], A

	mov		[tmp], 0				; Update threshold
	mov		A, [dahave]
	add		A, [ditave]
	adc		[tmp], 0	
	asr		[tmp]
	rrc		A
	mov		[threshold], A
	
	ret	

;-----------------------------------------------------------------------------
;  FUNCTION NAME: WaitTone
;
;  DESCRIPTION:
;     Waits for the beginning of detection of a tone
;
;  ARGUMENTS:
;     none.
;
;  RETURNS:
;     [period]	Time to detect the tone. 255 if no tone detected
;
;  SIDE EFFECTS:
;     none.
;
;-----------------------------------------------------------------------------

WaitTone:
	cmp		[_Counter5ms], 255
	jz		errorWaitTone
	mov		A, [ToneDetect]
	jz		WaitTone
	mov		A, [ToneDetect]
	jz		WaitTone
	mov		[period], [_Counter5ms]
	mov		[_Counter5ms], 0		
	call	DisplayTone
	ret	
errorWaitTone:
	mov		[period], 255
	mov		[_Counter5ms], 0		
	ret	


;-----------------------------------------------------------------------------
;  FUNCTION NAME: WaitSilence
;
;  DESCRIPTION:
;     Waits for the beginning of detection of a silence
;
;  ARGUMENTS:
;     none.
;
;  RETURNS:
;     [period]	Time to detect the tone. 255 if no silence detected
;
;  SIDE EFFECTS:
;     none.
;
;-----------------------------------------------------------------------------
WaitSilence:
	cmp		[_Counter5ms], 255
	jz		errorWaitSilence
	mov		A, [ToneDetect]
	jnz		WaitSilence
	mov		A, [ToneDetect]
	jnz		WaitSilence
	mov		[period], [_Counter5ms]
	mov		[_Counter5ms], 0		
	call	DisplaySilence
	ret	
errorWaitSilence:
	mov		[period], 255
	mov		[_Counter5ms], 0		
	ret

;-----------------------------------------------------------------------------
;  FUNCTION NAME: DisplayCharacter
;
;  DESCRIPTION:
; 	Display received character on first line of the display.
;	After the display is filled, the line is shifted to the left
;
;  ARGUMENTS:
;     [character]	Code of the character to display
;
;  RETURNS:
;     none.
;
;  SIDE EFFECTS:
;     none.
;
;-----------------------------------------------------------------------------
DisplayCharacter:
   	cmp		[column], 16			; Filled display ?	
   	jz		shiftDisplay	
   	
   	mov 	X, [column]				; Display character in current column and advance
   	mov		A, [character]			; column counter
   	mov		[X+linebuffer], A
   	inc		X
	inc		[column]
   	jmp		showDisplay
shiftDisplay:						; Shifts display to the left one position
	mov		X, 0
loopShiftDisplay:	
	inc		X
	mov		A, [X+linebuffer]
	dec		X
	mov		[X+linebuffer], A
	inc		X
	mov		A, X
	cmp		A, 15
	jnz		loopShiftDisplay
endshiftDisplay:	
	mov		X, 15
   	mov		A, [character]
   	mov		[X+linebuffer], A
showDisplay:   	
   	mov    	A, 0  					; First row and column          
	mov    	X, 0
   	lcall  	LCD_Position
   	
	mov     A, >linebuffer			; Display line buffer
	mov     X, <linebuffer
	lcall   LCD_PrString    
	
	call	DisplayStatusMsg

   	mov    	A, 1            
	mov    	X, 9
   	lcall  	LCD_Position
   	
	mov   	A, [dahave]     		; Compute WPM using table
	sub		A, 15
	cmp		A, 86
	jnc		errCalcWPM
	index	WPM_Table
	lcall  	LCD_PrHexByte        
	ret
errCalcWPM:
	mov		A, 0xff
	lcall  	LCD_PrHexByte        
	ret
	
;-----------------------------------------------------------------------------
;  FUNCTION NAME: TranslateCharacter
;
;  DESCRIPTION:
; 	Translates last received character using Morse code table.    
;
;  ARGUMENTS:
;     [code]		Index to the table
;
;  RETURNS:
;     [character]	Ascii code of the character
;
;  SIDE EFFECTS:
;     none.
;
;-----------------------------------------------------------------------------
	
TranslateCharacter:
	cmp		[code], 121
	jc		traslateCode
	mov		[character], '.'
	ret
traslateCode:	
	mov		A, [code]
	index	MorseCharTraslateTable
	mov		[character], A
	ret	

;-----------------------------------------------------------------------------
; WPM table as a function of DASH average period
;-----------------------------------------------------------------------------
WPM_Table:
	DB	0x48	; 15
	DB	0x45	; 16
	DB	0x42	; 17
	DB	0x40	; 18
	DB	0x38	; 19
	DB	0x36	; 20
	DB	0x34	; 21
	DB	0x33	; 22
	DB	0x31	; 23
	DB	0x30	; 24
	DB	0x29	; 25
	DB	0x28	; 26
	DB	0x27	; 27
	DB	0x26	; 28
	DB	0x25	; 29
	DB	0x24	; 30
	DB	0x23	; 31
	DB	0x22	; 32
	DB	0x22	; 33
	DB	0x21	; 34
	DB	0x21	; 35
	DB	0x20	; 36
	DB	0x19	; 37
	DB	0x19	; 38
	DB	0x18	; 39
	DB	0x18	; 40
	DB	0x18	; 41
	DB	0x17	; 42
	DB	0x17	; 43
	DB	0x16	; 44
	DB	0x16	; 45
	DB	0x16	; 46
	DB	0x15	; 47
	DB	0x15	; 48
	DB	0x15	; 49
	DB	0x14	; 50
	DB	0x14	; 51
	DB	0x14	; 52
	DB	0x14	; 53
	DB	0x13	; 54
	DB	0x13	; 55
	DB	0x13	; 56
	DB	0x13	; 57
	DB	0x12	; 58
	DB	0x12	; 59
	DB	0x12	; 60
	DB	0x12	; 61
	DB	0x12	; 62
	DB	0x11	; 63
	DB	0x11	; 64
	DB	0x11	; 65
	DB	0x11	; 66
	DB	0x11	; 67
	DB	0x11	; 68
	DB	0x10	; 69
	DB	0x10	; 70
	DB	0x10	; 71
	DB	0x10	; 72
	DB	0x10	; 73
	DB	0x10	; 74
	DB	0x10	; 75
	DB	0x09	; 76
	DB	0x09	; 77
	DB	0x09	; 78
	DB	0x09	; 79
	DB	0x09	; 80
	DB	0x09	; 81
	DB	0x09	; 82
	DB	0x09	; 83
	DB	0x09	; 84
	DB	0x08	; 85
	DB	0x08	; 86
	DB	0x08	; 87
	DB	0x08	; 88	
	DB	0x08	; 89
	DB	0x08	; 90
	DB	0x08	; 91
	DB	0x08	; 92
	DB	0x08	; 93
	DB	0x08	; 94
	DB	0x08	; 95
	DB	0x08	; 96
	DB	0x07	; 97
	DB	0x07	; 98
	DB	0x07	; 99
	DB	0x07	; 100
				
;-----------------------------------------------------------------------------
; Morse translation table
;-----------------------------------------------------------------------------
MorseCharTraslateTable:
	DB '.'
	DB '.' 	 
	DB 'T'	; -
	DB 'E'	; .
	DB 'M'	; --
	DB 'A'	; .-
	DB 'N'	; -.
	DB 'I'	; ..
	DB 'O'	; ---
	DB 'W'	; .--
	DB 'K'	; -.-
	DB 'U'	; ..-
	DB 'G'	; --. 
	DB 'R'	; .-.
	DB 'D'	; -..
	DB 'S'	; ...
	DB '.'	; ----
	DB 'J'	; .---
	DB 'Y'	; -.--
	DB 0xf5	; ..-- 
	DB 'Q'	; --.-
	DB 0xe1	; .-.- 	
	DB 'X'	; -..-
	DB 'V'	; ...-
	DB 0xef	; ---. 
	DB 'P'	; .--.
	DB 'C'	; -.-.
	DB 'F'	; ..-.
	DB 'Z'	; --..
	DB 'L'	; .-..
	DB 'B'	; -...
	DB 'H'	; ....
	DB '0'	; -----
	DB '1'	; .----
	DB '.'	; -.---
	DB '2'	; ..---
	DB 0xee	; --.--	
	DB '.'	; .-.--
	DB '.'	; -..--
	DB '3'	; ...--
	DB '.'	; ---.-
	DB 'A'	; .--.- 
	DB 0x7e	; -.-.- Starting signal
	DB '.'	; ..-.-
	DB '.'	; --..-
	DB '.'	; .-..-
	DB '='	; -...-
	DB '4'	; ....-
	DB '9'	; ----.
	DB '.'	; .---.
	DB '('	; -.--.
	DB '.'	; ..--.
	DB '.'	; --.-.
	DB '+'	; .-.-. End of message
	DB '/'	; -..-.
	DB '*'	; ...-. Understood
	DB '8'	; ---..
	DB '.'	; .--..
	DB '.'	; -.-..
	DB 'E'	; ..-..	
	DB '7'	; --...
	DB 0x7c	; .-... Wait
	DB '6'	; -....
	DB '5'	; .....
	DB '.'	; ------
	DB '.'	; .-----
	DB '.'	; -.----
	DB '.'	; ..----
	DB '.'	; --.---
	DB '.'	; .-.---
	DB '.'	; -..---
	DB '.'	; ...---
	DB '.'	; ---.--
	DB '.'	; .--.--
	DB '.'	; -.-.--
	DB '.'	; ..-.--
	DB ','	; --..--	','
	DB '.'	; .-..--
	DB '.'	; -...--
	DB '.'	; ....--
	DB '.'	; ----.-
	DB '.'	; .---.-
	DB ')'	; -.--.-
	DB '_'	; ..--.-
	DB '.'	; --.-.-
	DB '.'	; .-.-.-	'.'
	DB '.'	; -..-.-
	DB 0xff	; ...-.- End of work
	DB '.'	; ---..-
	DB '.'	; .--..-
	DB '.'	; -.-..-
	DB '.'	; ..-..-
	DB '.'	; --...-
	DB '.'	; .-...-
	DB '-'	; -....-
	DB '.'	; .....-
	DB '.'	; -----.
	DB '.'	; .----.
	DB '.'	; -.---.
	DB '.'	; ..---.
	DB '.'	; --.--.
	DB '.'	; .-.--.
	DB '.'	; -..--.
	DB '.'	; ...--.
	DB '.'	; ---.-.
	DB '.'	; .--.-.
	DB ';'	; -.-.-.
	DB '.'	; ..-.-.
	DB '.'	; --..-.
	DB '"'	; .-..-.
	DB '.'	; -...-.
	DB '.'	; ....-.
	DB '.'	; ----..
	DB '.'	; .---..
	DB '.'	; -.--..
	DB '?'	; ..--..	
	DB '.'	; --.-..
	DB 0xf0	; .-.-.. Paragraph
	DB '.'	; -..-..
	DB '.'	; ...-..
	DB ':'	; ---...
	DB '.'	; .--...

