;*****************************************************************************/
;
;	PROJECT:	PSOC MORSE DECODER
; 	FILE NAME: 	TONEDETECTOR.ASM
; 	AUTHOR:		Melchor Varela
;
; 	DESCRIPTION
;
;	Tone detector based on a quadrature correlator. The correlator correlates
; 	the incoming signal with two quadrature signals which correspond to the 
;	sign function of sine and cosine signals of a reference 800Hz sinusoid.
;	Besides, it performs tone debouncing and provides the 5ms timing reference. 
;
; 	HISTORY
;
;	NAME   	DATE		REMARKS
;
; 	MVM		11OCT2003	Creation
;	ASCII	30OCT2008	Migration to CY8C27442 , debugging
;
;*****************************************************************************/
;*****************************************************************************/
;
;	$Workfile:   tonedetector.asm  $
;
;	$Revision:   1.3  $
;
;*****************************************************************************/
include "m8c.inc"
include "multiply.inc"

;-----------------------------------------------------------------------------
;  Exported interfaces
;-----------------------------------------------------------------------------
export  TONEDET_Process
export  _TONEDET_Setup

;-----------------------------------------------------------------------------
;  Macros
;-----------------------------------------------------------------------------

;-----------------------------------------------------------------------------
;  Defines
;-----------------------------------------------------------------------------
SAMPLES_TO_PROCESS_N: 	equ 32
LOW_BYTE:  				equ 1
HIGH_BYTE: 				equ 0
ENERGY_THRESHOLD:		equ	5	; Threshold energy level to accept tone 		
DEBOUNCE_THRESHOLD:		equ	2	; Threshold debounce counter value to validate tone

area bss(RAM) 
;-----------------------------------------------------------------------------
;  Public data: 
;-----------------------------------------------------------------------------
export 	ToneDetect				
export	ToneLevel
export	_Counter5ms

	ToneDetect:		BLK	1		; 1 when tone detected. 0 if no tone				
	ToneLevel:		BLK	1		; Tone level
	_Counter5ms:		BLK	1		; 5ms timing reference

;-----------------------------------------------------------------------------
;  Private data: 
;-----------------------------------------------------------------------------
  	adc_sample:  	BLK 2		; Sample value
  	samples_cnt:  	BLK 1		; Samples counter
  	sin_acc:   		BLK 2		; Sine correlator accumulator
  	cos_acc:   		BLK 2		; Cosine correlator accumulator
  	energy:  		BLK 4		; To store energy computation result
  	tmp: 			BLK 4		; Temporary variable for multiply operations
  	debounceCnt:	BLK	1		; Debounce counter

	  
area text(ROM,REL)

;-----------------------------------------------------------------------------
;  FUNCTION NAME: TONEDET_Setup
;
;  DESCRIPTION:
;     Initialises tone detector variables.
;
;  ARGUMENTS:
;     none.
;
;  RETURNS:
;     none.
;
;  SIDE EFFECTS:
;     none.
;
;-----------------------------------------------------------------------------
_TONEDET_Setup:
	mov 	[samples_cnt], 0
 	mov		[debounceCnt], 0
 	mov		[ToneDetect], 0
 	mov		[ToneLevel], 0
    mov		[sin_acc+0], 0
    mov		[sin_acc+1], 0
    mov		[cos_acc+0], 0
    mov		[cos_acc+1], 0
	ret

;-----------------------------------------------------------------------------
;  FUNCTION NAME: TONEDET_Process
;
;  DESCRIPTION:
;     Tone detector.
;
;  ARGUMENTS:
;     A : signed 8 bit sample
;
;  RETURNS:
; 	  ToneDetect : 1 tone detected. 0 no tone.
;	  ToneLevel  : Tone magnitude
;	  Counter5ms : Counter unconditionally incremented every 5ms
;
;  SIDE EFFECTS:
;     none.
;
;-----------------------------------------------------------------------------
;	OPERATION
;
;	Tone detector based on a quadrature correlator. The correlator correlates
; 	the incoming signal with two quadrature signals which correspond to the 
;	sign function of sine and cosine signals of a reference 800Hz sinusoid.
; 	After processing N samples, the quadrature components are squared to eliminate
;	any signal-phase dependence.
;
;	The N value has been choosen in order to have a detection time of about 5ms.
;	The sampling frequency has been selected in order to have an integer number
; 	of signal cycles and samples. Thank to this the sign function can be obtained
;	from the samples counter: four samples positive and four negative.
;
;	Fsample = 6465 sps
;	N = 32 
;	Tdetect: 4.95ms
;	Fdetect = 808Hz
;	4 signal cycles in N. One cycle is 8 samples.
;	
;	Besides, it performs tone debouncing and provides the 5ms timing reference. 
;
;*****************************************************************************/
TONEDET_Process:
    										; Multiply input by 2 to extend dynamic range 
    asl 	A
    mov 	[(adc_sample + LOW_BYTE)], a
    mov 	[(adc_sample + HIGH_BYTE)], 0
    jnc 	isPositive
    mov 	[(adc_sample + HIGH_BYTE)], -1
    
    asl 	[(adc_sample + LOW_BYTE)]
    rlc 	[(adc_sample + HIGH_BYTE)]
isPositive:
  
; SIN correlator 
	mov		A, [samples_cnt]
	and		A, 0x4
	
	jnz		negCorrSin
    mov 	A, [(adc_sample+1)]    
    add 	[(sin_acc+1)], A
    mov	 	A, [(adc_sample+0)]
    adc 	[(sin_acc+0)], A		  
	jmp		endCorrSin
negCorrSin:	
    mov 	A, [(adc_sample+1)]
    sub 	[(sin_acc+1)], A
    mov 	A, [(adc_sample+0)]
    sbb 	[(sin_acc+0)], A
endCorrSin:

; COS correlator
	mov		A, [samples_cnt]
	add		A, 2				; 90
	and		A, 0x4
	
	jnz		negCorrCos
    mov 	A, [(adc_sample+1)]    
    add 	[(cos_acc+1)], A
    mov	 	A, [(adc_sample+0)]
    adc 	[(cos_acc+0)], A		  
	jmp		endCorrCos
negCorrCos:	
    mov 	A, [(adc_sample+1)]
    sub 	[(cos_acc+1)], A
    mov 	A, [(adc_sample+0)]
    sbb 	[(cos_acc+0)], A
endCorrCos:
         
    inc 	[samples_cnt] 
    cmp 	[samples_cnt], SAMPLES_TO_PROCESS_N
    jnz 	exitToneDetect
								
								; Processed N samples
	inc		[_Counter5ms]		; Update 5ms timer reference
    
								; Compute tone energy
	Multiply32s_16s_16s tmp, cos_acc, cos_acc
	Multiply32s_16s_16s energy, sin_acc, sin_acc
  	mov 	A, [(tmp+3)]
  	add 	[(energy+3)], A
  	mov 	A, [(tmp+2)]
  	adc 	[(energy+2)], A
  	mov 	A, [(tmp+1)]
  	adc 	[(energy+1)], A
  	mov 	A, [(tmp+0)]
  	adc 	[(energy+0)], A

    mov 	[samples_cnt], 0
    
    mov		[sin_acc+0], 0
    mov		[sin_acc+1], 0
    mov		[cos_acc+0], 0
    mov		[cos_acc+1], 0

	mov		[ToneLevel], [energy+1]
	
								; Tone debouncing  
	cmp		[energy+1], ENERGY_THRESHOLD
	jc		noTone
	cmp		[debounceCnt], DEBOUNCE_THRESHOLD
	jnz		incDebounce
	mov		[ToneDetect], 1
	jmp		toneProcessingEnd
incDebounce:
	inc		[debounceCnt]
	jmp	  	toneProcessingEnd      
noTone:        
	cmp		[debounceCnt], 0
	jnz		decDebounce
	mov		[ToneDetect], 0
	jmp		toneProcessingEnd
decDebounce:
	dec		[debounceCnt]
    
toneProcessingEnd:    
    
exitToneDetect:  
	ret


